@@ -11,6 +11,8 @@ import logging
 import os
 from pathlib import Path, PureWindowsPath
 from typing import Union, List, Tuple
+import re
+import datetime
 
 import numpy as np
 import pandas as pd
@@ -32,12 +34,9 @@ DATA_TYPES = ['Integer', 'Real', 'String']
 NUMERIC_TYPES = DATA_TYPES[0:2]
 NON_NUMERIC_TYPES = DATA_TYPES[-1]
 
-# Message for NotImplementedErrors.
-NIE_MSG = ('This method is either not complete or untested. We appreciate '
-           'contributions, so if you would like to complete and test this '
-           'method, please read contributing.md. If there is commented out '
-           'code, you can uncomment it and re-install esa from source at '
-           'your own risk.')
+# RequestBuildDate uses Delphi conventions, which counts days since
+# Dec. 30th, 1899.
+DAY_0 = datetime.date(year=1899, month=12, day=30)
 
 
 # noinspection PyPep8Naming
@@ -47,11 +46,11 @@ class SAW(object):
     # Class level property defining the fields which will be returned
     # for different ObjectTypes by the get_power_flow_results method.
     POWER_FLOW_FIELDS = {
-        'bus': ['BusNum', 'BusName', 'AreaNum', 'BusPUVolt', 'BusAngle', 'BusNetMW',
+        'bus': ['BusNum', 'BusName', 'BusPUVolt', 'BusAngle', 'BusNetMW',
                 'BusNetMVR'],
-        'gen': ['BusNum', 'GenID', 'AreaNum', 'GenMW', 'GenMVR'],
-        'load': ['BusNum', 'LoadID', 'AreaNum', 'LoadMW', 'LoadMVR'],
-        'shunt': ['BusNum', 'ShuntID', 'AreaNum', 'ShuntMW', 'ShuntMVR'],
+        'gen': ['BusNum', 'GenID', 'GenMW', 'GenMVR'],
+        'load': ['BusNum', 'LoadID', 'LoadMW', 'LoadMVR'],
+        'shunt': ['BusNum', 'ShuntID', 'ShuntMW', 'ShuntMVR'],
         'branch': ['BusNum', 'BusNum:1', 'LineCircuit', 'LineMW',
                    'LineMW:1', 'LineMVR', 'LineMVR:1']
     }
@@ -61,6 +60,9 @@ class SAW(object):
         ['key_field', 'internal_field_name', 'field_data_type', 'description',
          'display_name']
 
+    # Older versions of Simulator omit the "display name" field.
+    FIELD_LIST_COLUMNS_OLD = FIELD_LIST_COLUMNS[0:-1]
+
     # Class level property defining columns used for
     # GetSpecificFieldList method.
     SPECIFIC_FIELD_LIST_COLUMNS = \
@@ -141,6 +143,10 @@ class SAW(object):
         # Open the case.
         self.OpenCase(FileName=FileName)
 
+        # Get the version number and the build date
+        version_string, self.build_date = self.get_version_and_builddate()
+        self.version = int(re.search(r'\d+', version_string).group(0))
+
         # Look up and cache field listing and key fields for the given
         # object types in object_field_lookup.
         self._object_fields = dict()
@@ -422,7 +428,8 @@ class SAW(object):
         # Return a listing of the internal field name.
         return key_field_df['internal_field_name'].tolist()
 
-    def get_power_flow_results(self, ObjectType: str) -> \
+    def get_power_flow_results(self, ObjectType: str, additional_fields: Union[
+        None, List[str]] = None) -> \
             Union[None, pd.DataFrame]:
         """Get the power flow results from SimAuto server.
 
@@ -430,6 +437,9 @@ class SAW(object):
             are the keys in the POWER_FLOW_FIELDS class attribute (case
             insensitive).
 
+        :param additional_fields: Pass a list of field names to extend the
+            default attributes in the POWER_FLOW_FIELDS.
+
         :returns: Pandas DataFrame with the corresponding results, or
             None if the given ObjectType is not present in the model.
 
@@ -439,6 +449,8 @@ class SAW(object):
         # Get the listing of fields for this object type.
         try:
             field_list = self.POWER_FLOW_FIELDS[object_type]
+            if additional_fields:
+                field_list += additional_fields
         except KeyError:
             raise ValueError('Unsupported ObjectType for power flow results, '
                              '{}.'.format(ObjectType))
@@ -446,6 +458,12 @@ class SAW(object):
         return self.GetParametersMultipleElement(ObjectType=object_type,
                                                  ParamList=field_list)
 
+    def get_version_and_builddate(self) -> tuple:
+        return self._call_simauto("GetParametersSingleElement",
+                                  "PowerWorldSession",
+                                  convert_list_to_variant(["Version",
+                                                           "ExeBuildDate"]), convert_list_to_variant(["", ""]))
+
     def identify_numeric_fields(self, ObjectType: str,
                                 fields: Union[List, np.ndarray]) -> \
             np.ndarray:
@@ -534,6 +552,27 @@ class SAW(object):
                                  'not a valid path!'.format(property_value))
 
         # Set the property.
+        try:
+            self._set_simauto_property(property_name=property_name,
+                                       property_value=property_value)
+        except AttributeError as e:
+            if property_name == 'UIVisible':
+                self.log.warning(
+                    'UIVisible attribute could not be set. Note this SimAuto '
+                    'property was not introduced until Simulator version 20. '
+                    'Check your version with the get_simulator_version method.'
+                )
+            else:
+                raise e from None
+
+    def _set_simauto_property(self, property_name, property_value):
+        """Helper to just set a property name and value. Primary purpose
+        of breaking things out this way is for testing.
+
+        :param property_name: Name of the property to set, e.g.
+            UIVisible.
+        :param property_value: Value to set the property to, e.g. False.
+        """
         setattr(self._pwcom, property_name, property_value)
 
     def update_ui(self) -> None:
@@ -714,9 +753,9 @@ class SAW(object):
         :param copy: Whether or not to return a copy of the DataFrame.
             You may want a copy if you plan to make any modifications.
 
-        :returns: Pandas DataFrame with columns 'key_field,'
-            'internal_field_name,' 'field_data_type,' 'description,'
-            and 'display_name.'
+        :returns: Pandas DataFrame with columns from either
+            SAW.FIELD_LIST_COLUMNS or SAW.FIELD_LIST_COLUMNS_OLD,
+            depending on the version of PowerWorld Simulator being used.
 
         `PowerWorld Documentation
         <https://www.powerworld.com/WebHelp/#MainDocumentation_HTML/GetFieldList_Function.htm>`__
@@ -729,10 +768,39 @@ class SAW(object):
             output = self._object_fields[object_type]
         except KeyError:
             # We haven't looked up fields for this object yet.
-            # Call SimAuto, and place results into a DataFrame.
+            # Call SimAuto.
             result = self._call_simauto('GetFieldList', ObjectType)
-            output = pd.DataFrame(np.array(result),
-                                  columns=self.FIELD_LIST_COLUMNS)
+
+            # Place result in a numpy array.
+            result_arr = np.array(result)
+
+            # Attempt to map results into a DataFrame using
+            # FIELD_LIST_COLUMNS. If that fails, use
+            # FIELD_LIST_COLUMNS_OLD.
+            try:
+                output = pd.DataFrame(result_arr,
+                                      columns=self.FIELD_LIST_COLUMNS)
+            except ValueError as e:
+                # We may be dealing with the older convention.
+                # The value error should read something like:
+                # "Shape of passed values is (259, 4), indices imply (259, 5)"
+                # Confirm via regular expressions.
+                exp_base = r'\([0-9]+,\s'
+                exp_end = r'{}\)'
+                # Get number of columns for new/old lists.
+                nf_old = len(self.FIELD_LIST_COLUMNS_OLD)
+                nf_new = len(self.FIELD_LIST_COLUMNS)
+                # Search the error's arguments.
+                r1 = re.search(exp_base + exp_end.format(nf_old), e.args[0])
+                r2 = re.search(exp_base + exp_end.format(nf_new), e.args[0])
+
+                # Both results should match, i.e., not be None.
+                if (r1 is None) or (r2 is None):
+                    raise e
+
+                # If we made it here, use the older columns.
+                output = pd.DataFrame(result_arr,
+                                      columns=self.FIELD_LIST_COLUMNS_OLD)
 
             # While it appears PowerWorld gives us the list sorted by
             # internal_field_name, let's make sure it's always sorted.
@@ -1548,6 +1616,10 @@ class SAW(object):
         process. The property returns an integer value that represents a
         date. This information is useful for verifying the release
         version of the executable.
+
+        After contacting PowerWorld, it seems the integer comes back
+        according to Delphi date conventions, which counts days since
+        December 30th, 1899.
         """
         return self._pwcom.RequestBuildDate
 
@@ -1559,7 +1631,14 @@ class SAW(object):
         while using SimAuto. Set this property through the
         ``set_simauto_property`` method.
         """
-        return self._pwcom.UIVisible
+        try:
+            return self._pwcom.UIVisible
+        except AttributeError:
+            self.log.warning(
+                'UIVisible attribute could not be accessed. Note this SimAuto '
+                'property was not introduced until Simulator version 20. '
+                'Check your version with the get_simulator_version method.')
+            return False
 
     ####################################################################
     # Private Methods
