@@ -34,17 +34,19 @@ import logging
 import os
 import tempfile
 import unittest
-from unittest.mock import patch, MagicMock, Mock
+from unittest.mock import patch, MagicMock, Mock, seal
+import datetime
 
 import numpy as np
 import pandas as pd
 
-from esa import SAW, COMError, PowerWorldError, CommandNotRespectedError, Error
+from esa import SAW, COMError, PowerWorldError, CommandNotRespectedError, \
+    Error
 from esa.saw import convert_to_windows_path
 
 # noinspection PyUnresolvedReferences
-from tests.constants import PATH_14, PATH_14_PWD, PATH_2000, PATH_2000_mod, \
-    PATH_9, THIS_DIR, LTC_AUX_FILE, DATA_DIR
+from tests.constants import PATH_14, PATH_14_PWD, PATH_2000, \
+    PATH_9, THIS_DIR, AREA_AUX_FILE, DATA_DIR, VERSION
 
 # Initialize the 14 bus SimAutoWrapper. Adding type hinting to make
 # development easier.
@@ -52,6 +54,14 @@ from tests.constants import PATH_14, PATH_14_PWD, PATH_2000, PATH_2000_mod, \
 saw_14 = None  # type: SAW
 
 
+def skip_if_version_below(version=21):
+    """Use this method to skip tests below a given PowerWorld Simulator
+    version.
+    """
+    if VERSION < version:
+        raise unittest.SkipTest('Simulator version < {}.'.format(version))
+
+
 # noinspection PyPep8Naming
 def setUpModule():
     """In order to allow us to split the test cases by functionality
@@ -103,10 +113,16 @@ class InitializationTestCase(unittest.TestCase):
         for f in ['bus', 'shunt']:
             df = my_saw_14._object_fields[f]
             self.assertIsInstance(df, pd.DataFrame)
-            self.assertSetEqual({'key_field', 'internal_field_name',
-                                 'field_data_type', 'description',
-                                 'display_name'},
-                                set(df.columns.to_numpy()))
+
+            cols = df.columns.to_numpy().tolist()
+            if len(cols) == len(my_saw_14.FIELD_LIST_COLUMNS):
+                self.assertEqual(cols, my_saw_14.FIELD_LIST_COLUMNS)
+            elif len(cols) == len(my_saw_14.FIELD_LIST_COLUMNS_OLD):
+                self.assertEqual(cols, my_saw_14.FIELD_LIST_COLUMNS_OLD)
+            else:
+                raise AssertionError(
+                    'Columns, {}, do not match either FIELD_LIST_COLUMNS or '
+                    'FIELD_LIST_COLUMNS_OLD.'.format(cols))
 
     def test_error_during_dispatch(self):
         """Ensure an exception is raised if dispatch fails."""
@@ -398,12 +414,22 @@ class GetKeyFieldListTestCase(unittest.TestCase):
         # Ensure this is NOT cached.
         self.assertNotIn('3wxformer', saw_14._object_key_fields)
 
+        # Key fields have changed for 3 winding transformers between
+        # versions.
+        if VERSION in [21, 22]:
+            expected = ['BusIdentifier', 'BusIdentifier:1', 'BusIdentifier:2',
+                        'LineCircuit']
+        elif VERSION == 17:
+            expected = ['BusName_NomVolt:4', 'BusNum3W:3', 'LineCircuit']
+        else:
+            raise NotImplementedError(
+                'We do not know when key fields for 3 winding transformers'
+                'changed, and have thus far only looked into PWS versions '
+                '17 and 21. Please update this test if you are running a '
+                'different version of Simulator.')
+
         # Ensure the list comes back correctly.
-        self.assertListEqual(
-            ['BusIdentifier', 'BusIdentifier:1', 'BusIdentifier:2',
-             'LineCircuit'],
-            saw_14.get_key_field_list('3WXFormer')
-        )
+        self.assertListEqual(expected, saw_14.get_key_field_list('3WXFormer'))
 
 
 class GetPowerFlowResultsTestCase(unittest.TestCase):
@@ -443,6 +469,12 @@ class GetPowerFlowResultsTestCase(unittest.TestCase):
         """There are no shunts in the IEEE 14 bus model."""
         self.assertIsNone(saw_14.get_power_flow_results('shunt'))
 
+    def test_with_additional_fields(self):
+        """Add additional fields to the result"""
+        result = saw_14.get_power_flow_results(ObjectType='Bus',
+                                               additional_fields=['AreaNum'])
+        self.assertTrue('AreaNum' in result.columns.values.tolist())
+
 
 class IdentifyNumericFieldsTestCase(unittest.TestCase):
     """Test identify_numeric_fields."""
@@ -450,12 +482,31 @@ class IdentifyNumericFieldsTestCase(unittest.TestCase):
     # noinspection PyMethodMayBeStatic
     def test_correct(self):
         # Intentionally make the fields out of alphabetical order.
-        fields = ['LineStatus', 'LockOut', 'LineR', 'LineX', 'BusNum']
-        np.testing.assert_array_equal(
-            saw_14.identify_numeric_fields(
-                ObjectType='Branch', fields=fields),
-            np.array([False, False, True, True, True])
-        )
+        if VERSION in [21, 22]:
+            fields = ['LineStatus', 'LockOut', 'LineR', 'LineX', 'BusNum']
+            expected = np.array([False, False, True, True, True])
+        elif VERSION == 17:
+            # The LockOut field is not present in version 17.
+            fields = ['LineStatus', 'LineR', 'LineX', 'BusNum']
+            expected = np.array([False, True, True, True])
+        else:
+            raise NotImplementedError(
+                'If you encounter this error, please update this test for the '
+                'version of PowerWorld Simulator that you are using. Thus far,'
+                ' it has only been tested with version 17 and 21.'
+            )
+
+        actual = \
+            saw_14.identify_numeric_fields(ObjectType='Branch', fields=fields)
+        np.testing.assert_array_equal(actual, expected)
+
+
+class GetVersionAndBuildDateTestCase(unittest.TestCase):
+    """Test get_version_and_builddate."""
+
+    # noinspection PyMethodMayBeStatic
+    def test_correct(self):
+        self.assertIsInstance(saw_14.get_version_and_builddate(), tuple)
 
 
 class SetSimAutoPropertyTestCase(unittest.TestCase):
@@ -496,8 +547,13 @@ class SetSimAutoPropertyTestCase(unittest.TestCase):
         self.assertTrue(p.UIVisible)
 
     def test_set_ui_visible_false(self):
-        self.saw.set_simauto_property('UIVisible', False)
-        self.assertFalse(self.saw._pwcom.UIVisible)
+        # UIVisible introduced in version 20.
+        if VERSION >= 20:
+            self.saw.set_simauto_property('UIVisible', False)
+            self.assertFalse(self.saw._pwcom.UIVisible)
+        else:
+            with self.assertLogs(logger=self.saw.log, level='WARN'):
+                self.saw.set_simauto_property('UIVisible', False)
 
     def test_set_ui_visible_bad_value(self):
         with self.assertRaisesRegex(
@@ -520,6 +576,23 @@ class SetSimAutoPropertyTestCase(unittest.TestCase):
             self.saw.set_simauto_property(property_name='junk',
                                           property_value='42')
 
+    def test_attr_error_for_ui_visible(self):
+        """Force the UIVisible attribute to throw an error.
+        """
+        with self.assertLogs(logger=self.saw.log, level='WARN'):
+            with patch.object(self.saw, '_set_simauto_property',
+                              side_effect=AttributeError):
+                self.saw.set_simauto_property('UIVisible', True)
+
+    def test_re_raise_attr_error(self):
+        """Ensure that a non UIVisible property correctly re-raises
+        the attribute error.
+        """
+        with patch.object(self.saw, '_set_simauto_property',
+                          side_effect=AttributeError('my special error')):
+            with self.assertRaisesRegex(AttributeError, 'my special error'):
+                self.saw.set_simauto_property('CreateIfNotFound', False)
+
 
 class UpdateUITestCase(unittest.TestCase):
     """Test the update_ui method. To avoid conflicts with
@@ -891,8 +964,29 @@ class GetFieldListTestCase(unittest.TestCase):
     def check_field_list(self, field_list):
         """Helper to check a returned field list DataFrame."""
         self.assertIsInstance(field_list, pd.DataFrame)
-        self.assertEqual(saw_14.FIELD_LIST_COLUMNS,
-                         field_list.columns.to_numpy().tolist())
+        actual = field_list.columns.to_numpy().tolist()
+
+        # Check FIELD_LIST_COLUMNS first, then check
+        # FIELD_LIST_COLUMNS_OLD
+        new = False
+        old = False
+
+        try:
+            self.assertEqual(saw_14.FIELD_LIST_COLUMNS, actual)
+        except AssertionError:
+            new = True
+
+        try:
+            self.assertEqual(saw_14.FIELD_LIST_COLUMNS_OLD, actual)
+        except AssertionError as e2:
+            old = True
+
+        if new and old:
+            raise AssertionError('Valid columns:\n{}\n{}\nReceived:{}'
+                                 .format(saw_14.FIELD_LIST_COLUMNS,
+                                         saw_14.FIELD_LIST_COLUMNS_OLD,
+                                         actual))
+
         pd.testing.assert_frame_equal(
             field_list, field_list.sort_values(by=['internal_field_name']))
 
@@ -962,6 +1056,42 @@ class GetFieldListTestCase(unittest.TestCase):
         field_list = saw_14.GetFieldList('shunt')
         self.check_field_list(field_list)
 
+    def test_switch_to_old_field_list(self):
+        """Ensure that switching to "FIELD_LIST_COLUMNS_OLD works as
+        it should.
+        """
+        # Patch _object_fields to force a key error.
+        with patch.object(saw_14, '_object_fields', new={'dict': 1}):
+            # Patch _call_simauto to give us data that matches the shape
+            # of the FIELD_LIST_COLUMNS_OLD.
+            out = [['x'] * len(saw_14.FIELD_LIST_COLUMNS_OLD) for _ in
+                   range(10)]
+
+            with patch.object(saw_14, '_call_simauto', return_value=out):
+                result = saw_14.GetFieldList('bus')
+
+        # Ensure the result is a DataFrame.
+        self.assertIsInstance(result, pd.DataFrame)
+
+        # Ensure the columns are the "OLD" ones.
+        self.assertListEqual(result.columns.tolist(),
+                             saw_14.FIELD_LIST_COLUMNS_OLD)
+
+        # Ensure the shape is right. Hard-code the 10.
+        self.assertEqual(result.shape,
+                         (10, len(saw_14.FIELD_LIST_COLUMNS_OLD)))
+
+    def test_df_value_error_not_from_old_list(self):
+        """Make sure ValueError gets re-raised."""
+        # Patch DataFrame creation to raise error.
+        with patch('pandas.DataFrame', side_effect=ValueError('stuff n th')):
+            # Patch _object fields so that SimAuto is called. Using a
+            # dict for 'new' is Brandon's hack to force a KeyError.
+            with patch.object(saw_14, '_object_fields', new={'dict': 1}):
+                # Ensure we get our DataFrame ValueError back.
+                with self.assertRaisesRegex(ValueError, 'stuff n th'):
+                    saw_14.GetFieldList('bus')
+
 
 class GetParametersMultipleElementTestCase(unittest.TestCase):
     """Test GetParametersMultipleElement"""
@@ -1219,6 +1349,7 @@ class ListOfDevicesTestCase(unittest.TestCase):
     def test_buses(self):
         """As the name implies, we should get 14 buses."""
         result = saw_14.ListOfDevices(ObjType="Bus")
+        # noinspection PyTypeChecker
         expected = pd.DataFrame(
             data=np.arange(1, 15, dtype=np.int64).reshape(14, 1),
             index=np.arange(0, 14),
@@ -1330,7 +1461,7 @@ class ProcessAuxFileTestCase(unittest.TestCase):
 
     @classmethod
     def setUpClass(cls) -> None:
-        cls.saw = SAW(FileName=PATH_2000_mod, early_bind=True, UIVisible=False)
+        cls.saw = SAW(FileName=PATH_2000, early_bind=True, UIVisible=False)
 
     @classmethod
     def tearDownClass(cls) -> None:
@@ -1339,28 +1470,28 @@ class ProcessAuxFileTestCase(unittest.TestCase):
 
     def test_process_aux_ltc_filter(self):
         # Process the aux file. Note the aux file has a named filter,
-        # "ltc_filter"
-        self.saw.ProcessAuxFile(FileName=LTC_AUX_FILE)
+        # "area_filter"
+        self.saw.ProcessAuxFile(FileName=AREA_AUX_FILE)
 
-        # Get branch key fields.
-        kf = self.saw.get_key_field_list('branch')
+        # Get bus key fields.
+        kf = self.saw.get_key_field_list('bus')
 
-        # Get transformer data, using the named filter.
-        ltc = self.saw.GetParametersMultipleElement(
-            ObjectType='branch', ParamList=kf, FilterName="ltc_filter"
+        # Get bus data, using the named filter.
+        area_bus = self.saw.GetParametersMultipleElement(
+            ObjectType='bus', ParamList=kf, FilterName="area_filter"
         )
 
         # Check that the filter worked.
-        branch = self.saw.GetParametersMultipleElement(
-            ObjectType='branch', ParamList=kf + ['LineXFType']
+        bus = self.saw.GetParametersMultipleElement(
+            ObjectType='bus', ParamList=kf + ['AreaName']
         )
 
-        # Ensure the branch data is not getting filtered.
-        self.assertNotEqual(ltc.shape[0], branch.shape[0])
+        # Ensure the bus data is not getting filtered.
+        self.assertNotEqual(area_bus.shape[0], bus.shape[0])
 
-        # Ensure the number of LTC transformers matches up.
-        ltc_expected = branch['LineXFType'] == 'LTC'
-        self.assertEqual(ltc_expected.sum(), ltc.shape[0])
+        # Ensure the number of buses in the area matches up.
+        bus_expected = bus['AreaName'] == 'East'
+        self.assertEqual(bus_expected.sum(), area_bus.shape[0])
 
 
 class RunScriptCommandTestCase(unittest.TestCase):
@@ -1639,9 +1770,10 @@ class SendToExcel(unittest.TestCase):
 
 
 ########################################################################
-# SimAuto Properties tests
+# Properties tests
 ########################################################################
 
+# noinspection PyStatementEffect
 class SimAutoPropertiesTestCase(unittest.TestCase):
     """Test the SimAuto attributes."""
 
@@ -1659,11 +1791,36 @@ class SimAutoPropertiesTestCase(unittest.TestCase):
         self.assertIsInstance(bd, int)
 
     def test_ui_visible(self):
-        self.assertFalse(saw_14.UIVisible)
+        # UIVisible introduced in version 20.
+        if VERSION >= 20:
+            self.assertFalse(saw_14.UIVisible)
+        else:
+            with self.assertLogs(logger=saw_14.log, level='WARN'):
+                saw_14.UIVisible
+
+    def test_attr_error_ui_visible(self):
+        """Patch UIVisible, and ensure we get a warning when it throws
+        an attribute error.
+        """
+        # Create a mock for saw_14_pwcom. This is necessary because
+        # patching attributes of the actual COM object doesn't work
+        # very well.
+        com_patch = MagicMock()
+        # seal the mock so that attempts to access non-existent
+        # attribute will result in an AttributeError.
+        seal(com_patch)
+
+        # Mock out _pwcom and get the UIVisible attribute.
+        with patch.object(saw_14, '_pwcom', new=com_patch):
+            with self.assertLogs(logger=saw_14.log, level='WARN'):
+                result = saw_14.UIVisible
+
+        self.assertFalse(result)
 
     def test_create_if_not_found(self):
         self.assertFalse(saw_14.CreateIfNotFound)
 
+
 ########################################################################
 # ScriptCommand helper tests
 ########################################################################
@@ -1693,6 +1850,13 @@ class OpenOneLineTestCase(unittest.TestCase):
     unless there is an error.
     """
 
+    @classmethod
+    def setUpClass(cls) -> None:
+        # For now, skip old versions of PowerWorld.
+        # TODO: remove this skip statement after saving the .pwd files
+        #   in different PowerWorld formats.
+        skip_if_version_below(21)
+
     def test_open_default(self):
         """Open the correct pwd file.
         """
@@ -1712,6 +1876,13 @@ class CloseOnelineTestCase(unittest.TestCase):
     unless there is an error.
     """
 
+    @classmethod
+    def setUpClass(cls) -> None:
+        # For now, skip old versions of PowerWorld.
+        # TODO: remove this skip statement after saving the .pwd files
+        #   in different PowerWorld formats.
+        skip_if_version_below(21)
+
     def test_close_default(self):
         """Close the last focused oneline diagram.
         """
